"""
"""

__author__ = 'Robert Ancell <bob27@users.sourceforge.net>'
__license__ = 'GNU General Public License Version 2'
__copyright__ = 'Copyright 2005-2006  Robert Ancell'

import sys
import os

import ui
import scene
import game
import chess_board
import ai

import pgn

class Config:
    """
    """
    
    __directory = None
    
    def __init__(self):
        """
        """
        self.__directory = os.path.expanduser('~/.glchess')
        
        # Create the directory if it does not exist
        if not os.path.exists(self.__directory):
            os.mkdir(self.__directory)
        else:
            assert(os.path.isdir(self.__directory))

    def getAutosavePath(self):
        """Get the path to the autosave file"""
        return self.__directory + '/autosave.pgn'

class ScenePlayer(game.ChessPlayer):
    
    __scene = None
    
    __pieces = None
    
    def __init__(self, g, scene):
        """
        """
        self.__scene = scene
        self.__pieces = {}
        game.ChessPlayer.__init__(self, '@scene')
        g.addSpectator(self)
        
    def onPieceMoved(self, piece, start, end):
        """Called by chess_board.ChessPlayer"""
        # Piece creation
        if start is None:
            pieceName = {chess_board.PAWN: 'pawn', chess_board.ROOK: 'rook', chess_board.KNIGHT: 'knight',
                         chess_board.BISHOP: 'bishop', chess_board.QUEEN: 'queen', chess_board.KING: 'king'}[piece.getType()]
                    
            chessSet = {chess_board.WHITE: 'white', chess_board.BLACK: 'black'}[piece.getColour()]

            self.__pieces[piece] = self.__scene.addChessPiece(chessSet, pieceName, end)
        
        # Piece deletion
        elif end is None:
            scenePiece = self.__pieces.pop(piece)
            self.__scene.removeChessPiece(scenePiece)
        
        # Piece movement
        else:
            scenePiece = self.__pieces[piece]
            scenePiece.move(end)
            
        # Redraw the scene
        self.__scene.redraw()

class HumanPlayer(game.ChessPlayer):
    """
    """
    
    __scene = None
    
    def __init__(self, scene, name):
        game.ChessPlayer.__init__(self, name)
        self.__scene = scene
        
    def readyToMove(self):
        self.__scene._currentPlayer = self

class Scene(scene.SceneGL, scene.SceneHumanInput, ui.Renderable):
    """
    """
    
    __game = None
    
    # The (potential) human players
    __whitePlayer = None
    __blackPlayer = None
    
    # The current player
    _currentPlayer = None
    
    __view = None
    
    def __init__(self, game, ui):
        """
        """
        self.__ui = ui
        
        # The game and (potential) human players
        self.__game = game
        self.__whitePlayer = HumanPlayer(self, 'White')
        self.__blackPlayer = HumanPlayer(self, 'Black')

        scene.SceneHumanInput.__init__(self)
        scene.SceneGL.__init__(self)
        
    def getWhite(self):
        """Get the white human player.
        
        Return the white human player.
        """
        return self.__whitePlayer
    
    def getBlack(self):
        """Get the black human player.
        
        Return the black human player.
        """
        return self.__blackPlayer

    def playerIsHuman(self):
        """Check if the current player is a human.
        
        Return True the current player is human else False.
        """
        return self._currentPlayer is self.__whitePlayer or \
               self._currentPlayer is self.__blackPlayer

    def squareIsFriendly(self, coord):
        """Check if a given square contains a friendly piece.
        
        Return True if this square contains a friendly piece.
        """
        owner = self.__game.getSquareOwner(coord)
        if owner is None:
            return False
        return owner is self._currentPlayer
    
    def canMove(self, start, end):
        """Check if a move is valid.
        
        'start' is the square to move from (tuple (file,rank)).
        'end' is the square to move to (tuple (file,rank)).
        """
        if self._currentPlayer is None:
            return False

        return self._currentPlayer.canMove((start,end)) # FIXME: Promotion type
    
    def moveHuman(self, start, end):
        """Called when a human player moves.
        
        'start' is the location to move from (tuple (file,rank)).
        'end' is the location to move to (tuple (file,rank)).
        """
        player = self._currentPlayer
        self._currentPlayer = None
        player.move((start, end, chess_board.QUEEN)) # FIXME: Promotion type

    def setView(self, view):
        """
        """
        self.__view = view

    def redraw(self):
        """Redraw the view"""
        if self.__view is not None:
            self.__view.redraw()

    def onRedraw(self):
        """Called by scene.SceneGL"""
        self.redraw()
    
    def onRender(self):
        """Called by ui.Renderable"""
        self.render()
    
    def onReshape(self, width, height):
        """Called by ui.Renderable"""
        self.reshape(width, height)
    
    def onSelect(self, x, y):
        """Called by ui.Renderable"""
        self.select(x, y)
    
    def onDeselect(self, x, y):
        """Called by ui.Renderable"""
        self.deselect(x, y)
    
    def onClose(self):
        """Called by ui.Renderable"""
        # Accept the close and kill the view
        self.__ui.removeView(self.__view)

class GLChessGame(game.ChessGame):
    """
    """
    
    __name = None
    
    __scene = None
    __scenePlayer = None
    
    def __init__(self, name, ui):
        """
        """
        self.__name = name
        
        game.ChessGame.__init__(self)
        
        self.__scene = Scene(self, ui)
        self.__scenePlayer = ScenePlayer(self, self.__scene)
        
    def setWhiteAsHuman(self, name):
        """Set the white player to be a human.
        
        'name' is the name of the human player.
        """
        white = self.__scene.getWhite()
        white.setName(name)
        self.setWhite(white)
        
    def setBlackAsHuman(self, name):
        """Set the black player to be a human.
                
        'name' is the name of the human player.
        """
        black = self.__scene.getBlack()
        black.setName(name)
        self.setBlack(black)
        
    def getName(self):
        """
        """
        return self.__name
    
    def getScene(self):
        """
        """
        return self.__scene
    
    def toPGN(self, pgnGame):
        """Write the properties of this game into a PGN game.
        
        'pgnGame' is the game to write into (pgn.PGNGame). All the tags should be unset.
        """
        white = self.getWhite()
        black = self.getBlack()
        
        pgnGame.setTag(pgnGame.PGN_TAG_EVENT, self.getName())
        pgnGame.setTag(pgnGame.PGN_TAG_WHITE, white.getName())
        pgnGame.setTag(pgnGame.PGN_TAG_BLACK, black.getName())
            
        whiteType = white.getType()
        if whiteType is not None:
            pgnGame.setTag('WhiteAI', str(whiteType))
        blackType = black.getType()
        if blackType is not None:
            pgnGame.setTag('BlackAI', str(blackType))
            
        moves = self.getSANMoves()
        lastMove = None
        for m in moves:
            if lastMove is not None:
                pgnGame.addMove(lastMove, m)
                lastMove = None
            else:
                lastMove = m
        if lastMove is not None:
            pgnGame.addMove(lastMove, None)

class UI(ui.gtkui.gtkui.GtkUI):
    """
    """
    
    __application = None
    
    def __init__(self, application):
        """
        """
        self.__application = application
        ui.gtkui.gtkui.GtkUI.__init__(self)

    def onGameStart(self, gameName, allowSpectators, whiteName, whiteType, blackName, blackType, moves = None):
        """Called by UI"""
        g = self.__application.addGame(gameName, whiteName, whiteType, blackName, blackType)
        print 'Starting game ' + gameName + ' between ' + whiteName + '(' + str(whiteType) + ') and ' + blackName + '(' + str(blackType) + ')'
        if moves:
            g.start(moves)
        else:
            g.start()
        
    def onGameLoad(self, path, returnResult):
        """Called by UI"""
        try:
            p = pgn.PGN(path, 1)
        except pgn.Error, e:
            self.reportError('Unable to open PGN file ' + path, str(e))
            return
        
        # Use the first game
        pgnGame = p[0]
        
        if returnResult is True:
            whiteAI = pgnGame.getTag('WhiteAI')
            blackAI = pgnGame.getTag('BlackAI')
            msg = ''
            if whiteAI and self.__application.getAIProfile(whiteAI) is None:
                msg += "AI '" + whiteAI + "' is not installed, white player is now human\n"
                whiteAI = None
            if blackAI and self.__application.getAIProfile(blackAI) is None:
                msg += "AI '" + blackAI + "' is not installed, black player is now human\n"
                blackAI = None

            self.reportGameLoaded(gameName = pgnGame.getTag(pgnGame.PGN_TAG_EVENT),
                                  whiteName = pgnGame.getTag(pgnGame.PGN_TAG_WHITE),
                                  blackName = pgnGame.getTag(pgnGame.PGN_TAG_BLACK),
                                  whiteAI = whiteAI, blackAI = blackAI,
                                  moves = pgnGame.getMoves())
                                  
            if len(msg) > 0:
                self.reportError('Game modified', msg[:-1])
        else:
            g = self.__application.addPGNGame(pgnGame)
            moves = pgnGame.getMoves()
            if moves:
                g.start(moves)
            else:
                g.start()

    def onGameJoin(self, localName, localType, serverHost, serverId, game):
        """Called by UI"""
        print 'Joining game ' + game + ' on server ' + serverHost + ':' + str(serverId) + ' as ' + localName + '(' + str(localType) + ')'
        
    def onGameSave(self, view, path):
        """Called by UI"""
        g = self.__application.getGame(view)
        try:
            f = file(path, 'w')
        except IOError, e:
            self.reportError('Unable to save PGN file ' + path, str(e))
            return
        
        pgnGame = pgn.PGNGame()
        g.toPGN(pgnGame)
            
        lines = pgnGame.getLines()
        for line in lines:
            f.write(line + '\n')
        f.write('\n')
        f.close()
        
    def onViewRemoved(self, view):
        """Called by UI"""
        self.__application.removeGame(self.__application.getGame(view))

    def onQuit(self):
        """Called by UI"""
        self.__application.quit()

class GLChessApplication:
    """
    """
    
    # The configuration
    __config = None

    # The glChess UI
    __ui = None
    
    # The AI types and players in use
    __aiProfiles = None
    __aiPlayers = None
    
    # The games present
    __games = None
    __gamesByView = None
    __viewsByGame = None
    
    def __init__(self):
        """Constructor for glChess application"""
        self.__aiProfiles = {}
        self.__aiPlayers = []
        self.__games = []
        self.__gamesByView = {}
        self.__viewsByGame = {}
        
        self.__config = Config()

        self.__ui = UI(self)

        self.__ui.addTimer(10, self.__tickAIs)
        
    def addAIProfile(self, name, profile):
        """Add a new AI profile into glChess.
        
        'name' is the name of the profile (string).
        'profile' is the profile to add (ai.AIProfile).
        """
        assert(self.__aiProfiles.has_key(name) is False)
        self.__aiProfiles[name] = profile
        self.__ui.addAIEngine(name)
        
    def getAIProfile(self, name):
        """Get an installed AI profile.
        
        'name' is the name of the profile to get (string).
        
        Return the profile (ai.AIProfile) or None if it does not exist.
        """
        try:
            return self.__aiProfiles[name]
        except KeyError:
            return None

    def addGame(self, name, whiteName, whiteType, blackName, blackType):
        """Add a chess game into glChess.
        
        'name' is the name of the game (string).
        'whiteName' is the name of the white player (string).
        'whiteType' is the AI profile to use for white (string) or None if white is human.
        'blackName' is the name of the black player (string).
        'blackType' is the AI profile to use for black (string) or None if black is human.
        
        Returns the game object. Use game.start() to start the game.
        """
        g = GLChessGame(name, self.__ui)
        
        scene = g.getScene()
        view = self.__ui.addView(name, scene)
        scene.setView(view)

        self.__games.append(g)
        self.__gamesByView[view] = g
        self.__viewsByGame[g] = view

        msg = ''
        if whiteType is None:
            g.setWhiteAsHuman(whiteName)
        else:
            player = self.__addAIPlayer(whiteName, whiteType)
            if player:
                g.setWhite(player)
            else:
                msg += "AI '" + whiteType + "' is not installed, white player is now human\n"
                g.setWhiteAsHuman(whiteName)

        if blackType is None:
            g.setBlackAsHuman(blackName)
        else:
            player = self.__addAIPlayer(blackName, blackType)
            if player:
                g.setBlack(player)
            else:
                msg += "AI '" + blackType + "' is not installed, black player is now human\n"
                g.setBlackAsHuman(blackName)
                
        if len(msg) > 0:
            self.__ui.reportError('Game modified', msg[:-1])

        return g
    
    def addPGNGame(self, pgnGame):
        """Add a PGN game.
        
        'pgnGame' is the game to add (pgn.PGNGame).
        
        Returns the game object. Use game.start() to start the game.
        """
        return self.addGame(pgnGame.getTag(pgnGame.PGN_TAG_EVENT),
                            pgnGame.getTag(pgnGame.PGN_TAG_WHITE),
                            pgnGame.getTag('WhiteAI'),
                            pgnGame.getTag(pgnGame.PGN_TAG_BLACK),
                            pgnGame.getTag('BlackAI'))

    def getGame(self, view):
        """
        """
        return self.__gamesByView[view]

    def removeGame(self, game):
        """
        """
        game.end()
        view = self.__viewsByGame[game]
        self.__gamesByView.pop(view)
        self.__viewsByGame.pop(game)
        
        self.__games.remove(game)
        
        self.__ui.removeView(view)

    def start(self):
        """Run glChess.
        
        This method does not return.
        """
        # Setup some default AI profiles
        ais = [('GNUchess', 'gnuchess'), ('Sjeng', 'sjeng'), ('Amy', 'Amy'), ('crafty', 'crafty'), ('Invalid', 'invalid')]
        for (name, executable) in ais:
            self.__detectAI(name, executable)
        
        # Load existing games
        self.__autoload()

        # Start UI (does not return)
        self.__ui.run()

    def quit(self):
        """Quit glChess"""
        # Save any open games
        self.__autosave()
        
        # End current games (will delete AIs etc)
        for game in self.__games:
            game.end()

        # Exit the application
        sys.exit()
        
    # Private methods
    
    def __detectAI(self, name, executable):
        """Add an AI profile if the given executable exists.
        
        'name' is a descriptive name for the AI (string).
        'executable' is the executable to the AI engine (string).
        """
        if os.system('which ' + executable) != 0:
            return
        print 'Detected AI: ' + name
        app.addAIProfile(name, ai.AIProfile(executable))
    
    def __tickAIs(self):
        """Tick the AI players in glChess"""
        for player in self.__aiPlayers:
            player.tick()
        return True
    
    def __autoload(self):
        """Restore games from the autosave file"""
        try:
            p = pgn.PGN(self.__config.getAutosavePath())
            os.unlink(self.__config.getAutosavePath())
            games = p[:]
        except IOError:
            print 'No autosave'
            games = []

        moves = {}
        for pgnGame in games:
            g = self.addPGNGame(pgnGame)
            g.start(pgnGame.getMoves())
    
    def __autosave(self):
        """Save any open games to the autosave file"""
        if len(self.__games) == 0:
            return
        
        f = file(self.__config.getAutosavePath(), 'a')
        for g in self.__games:
            pgnGame = pgn.PGNGame()
            g.toPGN(pgnGame)
            
            lines = pgnGame.getLines()
            for line in lines:
                f.write(line + '\n')
            f.write('\n')
        f.close()
        
    def __addAIPlayer(self, name, aiType):
        """Create an AI player.
        
        'name' is the name of the player (string).
        'aiType' is the type of AI to use (string).
        
        Returns the new AI player or None if no player of this type.
        """
        try:
            profile = self.__aiProfiles[aiType]
        except KeyError:
            return None
        player = ai.AIPlayer(name, profile)
        player.setType(aiType)
        self.__aiPlayers.append(player)
        return player

if __name__ == '__main__':
    app = GLChessApplication()
    app.start()
