"""
"""

__author__ = 'Robert Ancell <bob27@users.sourceforge.net>'
__license__ = 'GNU General Public License Version 2'
__copyright__ = 'Copyright 2005-2006  Robert Ancell'

from OpenGL.GL import *
from OpenGL.GLU import *

import builtin_models

SQUARE_WIDTH      = 10.0
BOARD_DEPTH       = 2.0
BOARD_BORDER      = 5.0
BOARD_CHAMFER     = 0.0
BOARD_INNER_WIDTH = (SQUARE_WIDTH * 8.0)
BOARD_OUTER_WIDTH = (BOARD_INNER_WIDTH + BOARD_BORDER * 2.0)
OFFSET            = (BOARD_OUTER_WIDTH * 0.5)

# Highlight types
HIGHLIGHT_SELECTED = 'selected'
HIGHLIGHT_CAN_MOVE = 'canMove'

class ChessSet:
    """
    """

    def drawPiece(self, pieceName, state):
        """Draw a piece.
        
        'pieceName' is the piece name (string).
        'state' is the piece state (string).
        """
        pass
    
class ChessPiece:
    """
    """
    
    __chessSet = None
    __name = None
    
    pos = (0,0)
    
    def __init__(self, chessSet, name):
        self.__chessSet = chessSet
        self.__name = name
        
    def __coordToLocation(self, coord):
        """
        """
        file = coord[0]
        rank = coord[1]
        x = BOARD_BORDER + float(file) * SQUARE_WIDTH + 0.5 * SQUARE_WIDTH
        z = -(BOARD_BORDER + float(rank) * SQUARE_WIDTH + 0.5 * SQUARE_WIDTH)
        
        return (x, 0.0, z)

    def move(self, coord):
        self.pos = self.__coordToLocation(coord)
    
    def draw(self, state = 'default'):
        self.__chessSet.drawPiece(self.__name, state)

class SceneHumanInput:
    """
    """

    # The selected square to move from
    __startSquare = None
    
    def __init__(self):
        """Constructor for a scene with human selectable components"""
        pass
    
    def playerIsHuman(self):
        """Check if the current player is a human.
        
        Return True the current player is human else False.
        """
        return False

    def squareIsFriendly(self, coord):
        """Check if a given square contains a friendly piece.
        
        Return True if this square contains a friendly piece.
        """
        return False
    
    def canMove(self, start, end):
        """Check if a move is valid.
        
        'start' is the square to move from (tuple (file,rank)).
        'end' is the square to move to (tuple (file,rank)).
        """
        return False
    
    def moveHuman(self, start, end):
        """Called when a human player moves.
        
        'start' is the location to move from (tuple (file,rank)).
        'end' is the location to move to (tuple (file,rank)).
        """
        pass
    
    def setBoardHighlight(self, coords):
        """Called when a human player changes the highlighted squares.
        
        'coords' is a list or tuple of co-ordinates to highlight.
                 The co-ordinates are a tuple in the form (file,rank).
                 If None the highlight will be cleared.
        """
        pass
    
    def select(self, x, y):
        """
        """
        # Only bother if the current player is human
        if self.playerIsHuman() is False:
            return
        
        # Get the selected square
        coord = self.getSquare(x, y)
        if coord is None:
            return
                
        # If this is a friendly piece then select it
        if self.squareIsFriendly(coord):
            self.__startSquare = coord
            
            # Highlight the squares that can be moved to
            highlights = {}
            for file in (0, 1, 2, 3, 4, 5, 6, 7):
                for rank in (0, 1, 2, 3, 4, 5, 6, 7):
                    if self.canMove(coord, (file,rank)):
                        highlights[(file,rank)] = HIGHLIGHT_CAN_MOVE
            highlights[coord] = HIGHLIGHT_SELECTED
            self.setBoardHighlight(highlights)

        else:
            # If we have already selected a start move try
            # and move to this square
            if self.__startSquare is not None:
                self.__move(self.__startSquare, coord)

        # Redraw the scene
        self.onRedraw()
        
        return coord

    def deselect(self, x, y):
        """
        """
        # Only bother if the current player is human
        if self.playerIsHuman() is False:
            return
        
        # Get the selected square
        coord = self.getSquare(x, y)
        if coord is None:
            return
        
        # Attempt to move here
        if self.__startSquare is not None:
            self.__move(self.__startSquare, coord)
        
        # Redraw the scene
        self.onRedraw()
        
        return coord
    
    def __move(self, start, end):
        """Attempt to make a move.
        
        ...
        """
        if self.canMove(start,end) is False:
            return
        self.__selectedSquare = None
        self.setBoardHighlight(None)
        self.moveHuman(start, end)

class SceneGL:
    """
    """
    
    __viewportWidth = 0
    __viewportHeight = 0
    __viewportAspect = 1.0
    
    __lightPos = None
    
    __currentRotation = 0.0
    
    __pieces = None    
    __chessSets = None
    
    __highlights = None

    def __init__(self):
        """Constructor for an OpenGL scene"""
        self.__lightPos = [100.0, 100.0, 100.0, 1.0]
        self.__pieces = []
        self.__highlights = {}
        
        self.__chessSets = {'white': builtin_models.WhiteBuiltinSet(), 'black': builtin_models.BlackBuiltinSet()}
        
    def onRedraw(self):
        """This method is called when the scene needs redrawing"""
        pass

    def addChessPiece(self, chessSet, name, coord):
        """Add a chess piece model into the scene.
        
        'chessSet' is the name of the chess set (string).
        'name' is the name of the piece (string).
        'coord' is the the chess board location of the piece (tuple, (file,rank)).
        """
        chessSet = self.__chessSets[chessSet]
        piece = ChessPiece(chessSet, name)
        self.__pieces.append(piece)

        # Set the location on the chess square
        piece.move(coord)
        
        # Redraw the scene
        self.onRedraw()
        
        return piece
    
    def removeChessPiece(self, piece):
        """Remove chess piece.
        
        'piece' is a chess piece instance as returned by addChessPiece().
        """
        self.__pieces.remove(piece)
        self.onRedraw()
        
    def setBoardHighlight(self, coords):
        """Highlight a square on the board.
        
        'coords' is a dictionary of highlight types keyed by square co-ordinates.
                 The co-ordinates are a tuple in the form (file,rank).
                 If None the highlight will be cleared.
        """
        if coords is None:
            self.__highlights = {}
        else:
            self.__highlights = coords.copy()
        self.onRedraw()
    
    def reshape(self, width, height):
        """Resize the viewport into the scene.
        
        'width' is the width of the viewport in pixels.
        'height' is the width of the viewport in pixels.
        """
        self.__viewportWidth = int(width)
        self.__viewportHeight = int(height)
        self.__viewportAspect = float(self.__viewportWidth) / float(self.__viewportHeight)

    def render(self):
        """Render the scene.
        
        This requires an OpenGL context.
        """
        # Set the projection for this scene
        self.__setViewport()
        
        # Do camera and board rotation/translation
        self.__transformCameraAndBoard()

        self.__drawBoard()
        self.__drawPieces()

    def getSquare(self, x, y):
        """Find the chess square at a given 2D location.
        
        'x' is the number of pixels from the left of the scene to select.
        'y' is the number of pixels from the bottom of the scene to select.
        
        This requires an OpenGL context.
        
        Return the co-ordinate as a tuple in the form (file,rank) or None if
        no square at this point.
        """
        # FIXME: Don't rely on this? It seems to get corrupt when multiple games are started
        viewport = glGetIntegerv(GL_VIEWPORT)

        # Don't render to screen, just select
        # Selection buffer is large in case we select multiple squares at once (it generates an exception)
        glSelectBuffer(20)
        glRenderMode(GL_SELECT)

        glInitNames()
        glPushName(0)

        # Create pixel picking region near cursor location
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPickMatrix(x, (float(viewport[3]) - y), 1.0, 1.0, viewport)
        gluPerspective(60.0, float(viewport[2]) / float(viewport[3]), 0.1, 1000)

        # Draw board
        self.__transformCameraAndBoard()
        self.__drawSquares()

        # Render and check for hits
        # Catch the exception in case we select more than we can fit in the selection buffer
        glFlush()
        try:
            records = glRenderMode(GL_RENDER)
        except GLerror:
            records = None

        # Get the first record and use this as the selected square
        coord = None
        if records is not None:
            for record in records:
                coord = record[2]
                break

        # Reset projection matrix
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(60.0, float(viewport[2]) / float(viewport[3]), 0.1, 1000)

        return coord

    def __setViewport(self):
        """Perform the projection matrix transformation for the current viewport"""
        glViewport(0, 0, self.__viewportWidth, self.__viewportHeight)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(60.0, self.__viewportAspect, 0.1, 1000)

    def __transformCamera(self):
        """Perform the camera matrix transformation"""
        gluLookAt(0.0, 100.0, 60.0,
                  0.0,   0.0,  0.0,
                  0.0,   1.0,  0.0)

    def __drawBoard(self):
        """Draw a chessboard"""
        glLightfv(GL_LIGHT0, GL_POSITION, self.__lightPos)
        glEnable(GL_LIGHTING)
        glEnable(GL_COLOR_MATERIAL)
        glEnable(GL_LIGHT0)
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_CULL_FACE)

        # Draw the border
        glColor3f(0.36, 0.21, 0.05)

        # Top
        verticies = [(BOARD_CHAMFER, BOARD_DEPTH, -BOARD_CHAMFER), (BOARD_BORDER, BOARD_DEPTH, -BOARD_BORDER),
                     (BOARD_CHAMFER, BOARD_DEPTH, -BOARD_OUTER_WIDTH + BOARD_CHAMFER), (BOARD_BORDER, BOARD_DEPTH, -BOARD_OUTER_WIDTH + BOARD_BORDER),
                     (BOARD_OUTER_WIDTH - BOARD_CHAMFER, BOARD_DEPTH, -BOARD_OUTER_WIDTH + BOARD_CHAMFER), (BOARD_OUTER_WIDTH - BOARD_BORDER, BOARD_DEPTH, -BOARD_OUTER_WIDTH + BOARD_BORDER),
                     (BOARD_OUTER_WIDTH - BOARD_CHAMFER, BOARD_DEPTH, -BOARD_CHAMFER), (BOARD_OUTER_WIDTH - BOARD_BORDER, BOARD_DEPTH, -BOARD_BORDER),
                     (BOARD_CHAMFER, BOARD_DEPTH, -BOARD_CHAMFER), (BOARD_BORDER, BOARD_DEPTH, -BOARD_BORDER)]

        glBegin(GL_QUAD_STRIP)
        glNormal3f(0.0, 1.0, 0.0)
        for v in verticies:
            glVertex3fv(v)
        glEnd()

        # Sides
        glBegin(GL_QUAD_STRIP)
        glNormal3f(-0.71, 0.0, 0.71)
        glVertex3f(0.0, 0.0, 0.0)
        glVertex3f(0.0, BOARD_DEPTH, 0.0)

        glNormal3f(-0.71, 0.0, -0.71)
        glVertex3f(0.0, 0.0, -BOARD_OUTER_WIDTH)
        glVertex3f(0.0, BOARD_DEPTH, -BOARD_OUTER_WIDTH)

        glNormal3f(0.71, 0.0, -0.71)
        glVertex3f(BOARD_OUTER_WIDTH, 0.0, -BOARD_OUTER_WIDTH)
        glVertex3f(BOARD_OUTER_WIDTH, BOARD_DEPTH, -BOARD_OUTER_WIDTH)

        glNormal3f(0.71, 0.0, 0.71)
        glVertex3f(BOARD_OUTER_WIDTH, 0.0, 0.0)
        glVertex3f(BOARD_OUTER_WIDTH, BOARD_DEPTH, 0.0)

        glVertex3f(0.0, 0.0, 0.0)
        glVertex3f(0.0, BOARD_DEPTH, 0.0)
        glEnd()

        # Draw the squares
        glBegin(GL_QUADS)
        for x in [0, 1, 2, 3, 4, 5, 6, 7]:
            for y in [0, 1, 2, 3, 4, 5, 6, 7]:
                isBlack = (x + (y % 2) + 1) % 2

                # Get the highlight type
                coord = (x,y)
                try:
                    highlight = self.__highlights[coord]
                except KeyError:
                    highlight = None
                
                if isBlack:
                    colour = {None: (0.0, 0.0, 0.0), HIGHLIGHT_SELECTED: (0.1, 0.5, 0.0), HIGHLIGHT_CAN_MOVE: (0.0, 0.0, 0.4)}[highlight]
                else:
                    colour = {None: (1.0, 1.0, 1.0), HIGHLIGHT_SELECTED: (0.2, 1.0, 0.0), HIGHLIGHT_CAN_MOVE: (0.2, 0.2, 0.8)}[highlight]

                x0 = BOARD_BORDER + (x * SQUARE_WIDTH)
                x1 = x0 + SQUARE_WIDTH
                z0 = BOARD_BORDER + (y * SQUARE_WIDTH)
                z1 = z0 + SQUARE_WIDTH

                glColor3fv(colour)
                glTexCoord2f(0.0, 0.0)
                glVertex3f(x0, BOARD_DEPTH, -z0)
                glTexCoord2f(1.0, 0.0)
                glVertex3f(x1, BOARD_DEPTH, -z0)
                glTexCoord2f(1.0, 1.0)
                glVertex3f(x1, BOARD_DEPTH, -z1)
                glTexCoord2f(0.0, 1.0)
                glVertex3f(x0, BOARD_DEPTH, -z1)
        glEnd()
        
    def __drawSquares(self):
        """Draw the board squares for picking"""

        # draw the floor squares
        for u in [0, 1, 2, 3, 4, 5, 6, 7]:
            glLoadName(u)

            for v in [0, 1, 2, 3, 4, 5, 6, 7]:
                glPushName(v)

                # Draw square
                glBegin(GL_QUADS)
                x0 = BOARD_BORDER + (u * SQUARE_WIDTH)
                x1 = x0 + SQUARE_WIDTH
                z0 = BOARD_BORDER + (v * SQUARE_WIDTH)
                z1 = z0 + SQUARE_WIDTH

                glVertex3f(x0, BOARD_DEPTH, -z0)
                glVertex3f(x1, BOARD_DEPTH, -z0)
                glVertex3f(x1, BOARD_DEPTH, -z1)
                glVertex3f(x0, BOARD_DEPTH, -z1)
                glEnd()

                glPopName()
        
    def __drawPieces(self):
        """Draw the pieces in the scene"""

        for piece in self.__pieces:
            glPushMatrix()
            glTranslatef(piece.pos[0], piece.pos[1], piece.pos[2])

            # Draw the model
            piece.draw()
            
            glPopMatrix()
        
    def __transformCameraAndBoard(self):
        """Do the camera and board transforms"""
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        # Place the camera
        self.__transformCamera()

        # Rotate the board
        glRotatef(self.__currentRotation, 0.0, 1.0, 0.0)

        # Offset board so centre is (0.0,0.0)
        glTranslatef(-OFFSET, 0.0, OFFSET)
